import collections
import functools
import logging
import time
import typing as tp

import math

from satella.coding import precondition
from .base import EmbeddedSubmetrics, RUNTIME, DEBUG

logger = logging.getLogger(__name__)


# shamelessly taken from http://code.activestate.com/recipes/511478-finding-the-percentile-of-the-values/)
def percentile(n: tp.List[float], percent: float) -> float:
    """
    Find the percentile of a list of values.

    :param n: - is a list of values. Note this MUST BE already sorted.
    :param percent: - a float value from 0.0 to 1.0.

    :return: the percentile of the values
    """
    k = (len(n) - 1) * percent
    f = math.floor(k)
    c = math.ceil(k)
    if f == c:
        return n[int(k)]
    d0 = n[int(f)] * (c - k)
    d1 = n[int(c)] * (k - f)
    return d0 + d1


class PercentileMetric(EmbeddedSubmetrics):
    """
    A metric that can register some values, sequentially, and then calculate percentiles from it

    :param last_calls: last calls to handle() to take into account
    :param percentiles: a sequence of percentiles to return in to_json
    """

    CLASS_NAME = 'percentile'

    def __init__(self, name, root_metric: 'Metric' = None, metric_level: str = None,
                 last_calls: int = 100, percentiles: tp.Sequence[float] = (0.5, 0.95), *args,
                 **kwargs):
        super().__init__(name, root_metric, metric_level, *args, last_calls=last_calls,
                         percentiles=percentiles, **kwargs)
        self.last_calls = last_calls
        self.calls_queue = collections.deque()
        self.percentiles = percentiles

    def _handle(self, time_taken: float, **labels) -> None:
        if labels or self.embedded_submetrics_enabled:
            return super()._handle(time_taken, **labels)
        if len(self.calls_queue) == self.last_calls:
            self.calls_queue.pop()
        self.calls_queue.appendleft(time_taken)

    def to_json(self) -> dict:
        if self.embedded_submetrics_enabled:
            return super().to_json()

        output = []
        sorted_calls = sorted(self.calls_queue)
        for p_val in self.percentiles:
            k = super().to_json()
            if not sorted_calls:
                k.update(quantile=p_val, _=0.0)
            else:
                k.update(quantile=p_val,
                         _=percentile(sorted_calls, p_val))
            output.append(k)
        return output

    @precondition(None, None, lambda x: x in (RUNTIME, DEBUG))
    def measure(self, include_exceptions: bool = True, logging_level: int = RUNTIME,
                value_getter: tp.Callable[[], float] = time.monotonic, **labels):
        """
        A decorator to measure a difference between some value after the method call
        and before it.

        By default, it will measure the execution time.

        Use like:

        >>> call_time = getMetric('root.metric_name.execution_time', 'percentile')
        >>> @call_time.measure()
        >>> def measure_my_execution(args):
        >>>     ...

        :param include_exceptions: whether to include exceptions
        :param logging_level: one of RUNTIME or DEBUG
        :param value_getter: a callable that takes no arguments and returns a float, which is
            the value
        :param labels: extra labels to call handle() with
        """

        def outer(fun):
            @functools.wraps(fun)
            def inner(*args, **kwargs):
                start_value = value_getter()
                excepted = None
                try:
                    return fun(*args, **kwargs)
                except Exception as e:
                    excepted = e
                finally:
                    value_taken = value_getter() - start_value
                    if excepted is not None and not include_exceptions:
                        raise e

                    self.handle(logging_level, value_taken, **labels)

                    if e is not None:
                        raise e

            return inner

        return outer



import typing as tp
import copy
from abc import abstractmethod, ABCMeta
from satella.json import JSONAble
from satella.coding import for_argument


DISABLED = 1
RUNTIME = 2
DEBUG = 3
INHERIT = 4


class Metric(JSONAble):
    """
    A base metric class
    """
    CLASS_NAME = 'base'

    def reset(self) -> None:
        """Delete all child metrics that this metric contains"""
        from satella.instrumentation import metrics
        if self.name == '':
            metrics.metrics = {}
        else:
            metrics.metrics = {k: v for k, v in metrics.metrics.items() if
                               not k.startswith(self.name + '.')}
            del metrics.metrics[self.name]
        self.children = []

    def __init__(self, name, root_metric: 'Metric' = None, metric_level: str = None):
        """When reimplementing the method, remember to pass kwargs here!"""
        self.name = name
        self.root_metric = root_metric
        self.level = metric_level or RUNTIME
        assert not (
                self.name == '' and self.level == INHERIT), 'Unable to set INHERIT for root metric!'
        self.children = []

    def __str__(self) -> str:
        return self.name

    def append_child(self, metric: 'Metric'):
        self.children.append(metric)

    def can_process_this_level(self, target_level: int) -> bool:
        metric = self
        while metric.level == INHERIT:
            # this is bound to terminate, since it is not possible to set metric_level of INHERIT on root
            metric = metric.root_metric
        return metric.level >= target_level

    def switch_level(self, level: int) -> None:
        assert not (self.name == '' and level == INHERIT), 'Unable to set INHERIT for root metric!'
        self.level = level

    def to_json(self) -> tp.Union[list, dict, str, int, float, None]:
        return {
            child.name[len(self.name) + 1 if len(self.name) > 0 else 0:]: child.to_json() for child
            in self.children
        }

    def _handle(self, *args, **kwargs) -> None:
        """
        Override me!
        """
        raise NotImplementedError('This is an abstract method!')

    def handle(self, level: int, *args, **kwargs) -> None:
        if self.can_process_this_level(level):
            return self._handle(*args, **kwargs)

    def debug(self, *args, **kwargs):
        self.handle(DEBUG, *args, **kwargs)

    def runtime(self, *args, **kwargs):
        self.handle(RUNTIME, *args, **kwargs)


class LeafMetric(Metric):
    """
    A metric capable of generating only leaf entries.

    You cannot hook up any children to a leaf metric.
    """
    def __init__(self, name, root_metric: 'Metric' = None, metric_level: str = None,
                 labels: tp.Optional[dict] = None):
        super().__init__(name, root_metric, metric_level)
        self.labels = labels or {}

    def to_json(self) -> dict:
        return copy.copy(self.labels)

    def append_child(self, metric: 'Metric'):
        raise TypeError('This metric cannot contain children!')


class EmbeddedSubmetrics(LeafMetric):
    """
    A metric that can optionally accept some labels in it's handle, and this will be counted as a
    separate metric.
    For example:

    >>> metric = getMetric('root.test.IntValue', 'int')
    >>> metric.handle(2, label='key')
    >>> metric.handle(3, label='value')
    >>> assert metric.to_json() == [{'label': 'key', '_': 2}, {'label': 'value', '_': 3}]

    If you try to inherit from it, refer to :py:class:`.simple.IntegerMetric` to see how to do it.
    And please pass all the arguments received from child class into this constructor, as this
    constructor actually stores them!
    Refer to :py:class:`.cps.ClicksPerTimeUnitMetric` on how to do that.
    """
    def __init__(self, name, root_metric: 'Metric' = None, metric_level: str = None,
                 labels: tp.Optional[dict] = None, *args, **kwargs):
        super().__init__(name, root_metric, metric_level, labels)
        self.args = args
        self.kwargs = kwargs
        self.embedded_submetrics_enabled = False        # to check for in children
        self.children_mapping = {}

    def _handle(self, value, **labels):
        key = tuple(sorted(labels.items()))
        if key:
            self.embedded_submetrics_enabled = True
        else:
            return

        if key in self.children_mapping:
            # noinspection PyProtectedMember
            self.children_mapping[key]._handle(value)
        else:
            clone = self.clone(labels)
            self.children_mapping[key] = clone
            self.children.append(clone)
            # noinspection PyProtectedMember
            self.children_mapping[key]._handle(value)

    def to_json(self) -> list:
        if self.embedded_submetrics_enabled:
            return [child.to_json() for child in self.children]
        else:
            return super().to_json()

    def clone(self, labels: dict) -> 'LeafMetric':
        """
        Return a fresh instance of this metric, with it's parent being set to this metric
        and having a particular set of labels, and being of level INHERIT.
        """

        return self.__class__(self.name, self, INHERIT, *self.args, labels=labels, **self.kwargs)



import typing as tp

from .base import EmbeddedSubmetrics


class SimpleMetric(EmbeddedSubmetrics):
    CLASS_NAME = 'string'
    CONSTRUCTOR = str

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.data = None

    def _handle(self, *args, **kwargs) -> None:
        if self.embedded_submetrics_enabled or kwargs:
            return super()._handle(*args, **kwargs)

        self.data = self.CONSTRUCTOR(args[0])

    def to_json(self) -> tp.Union[list, dict, str, int, float, None]:
        if self.embedded_submetrics_enabled:
            return super().to_json()
        p = super().to_json()
        p['_'] = self.data
        return p


class StringMetric(SimpleMetric):
    pass


class IntegerMetric(SimpleMetric):
    CLASS_NAME = 'int'
    CONSTRUCTOR = int


class FloatMetric(SimpleMetric):
    CLASS_NAME = 'float'
    CONSTRUCTOR = float



import logging
import io

from ..json import is_leaf_node

logger = logging.getLogger(__name__)


class RendererObject(io.StringIO):

    def render_node(self, tree, prefixes):
        if isinstance(tree, list):
            for elem in tree:
                self.render_node(elem, prefixes=prefixes)
            return

        if is_leaf_node(tree):
            self.write('_'.join(prefixes))
            self.write('{')
            main_value = tree.pop('_')
            labels = []
            for key, value in tree.items():
                value = repr(value).replace('\\', '\\\\').replace('"', '\"').replace("'", '"')
                labels.append('%s=%s' % (key, value))
            self.write(', '.join(labels))
            self.write('} %s\n' % (repr(main_value), ))

        else:
            for k, v in tree.items():
                self.render_node(v, prefixes=prefixes + [k])


def json_to_prometheus(tree) -> str:
    """
    Render the JSON in the form understandable by Prometheus

    :param tree: JSON returned by the root metric (or any metric for that instance).
    :return: a string output to present to Prometheus
    """
    obj = RendererObject()
    obj.render_node(tree, [])
    return obj.getvalue()



import collections
import time
import typing as tp
import logging

from .base import EmbeddedSubmetrics


logger = logging.getLogger(__name__)


class ClicksPerTimeUnitMetric(EmbeddedSubmetrics):
    """
    This tracks the amount of calls to handle() during the last time periods, as specified by time_unit_vectors
    (in seconds). You may specify multiple time periods as consequent entries in the list.

    By default (if you do not specify otherwise) this will track calls made during the last second.
    """
    CLASS_NAME = 'cps'

    def __init__(self, *args, time_unit_vectors: tp.Optional[tp.List[float]] = None, **kwargs):
        super().__init__(*args, time_unit_vectors=time_unit_vectors, **kwargs)
        time_unit_vectors = time_unit_vectors or [1]
        self.last_clicks = collections.deque()
        self.cutoff_period = max(time_unit_vectors)
        self.time_unit_vectors = time_unit_vectors

    def _handle(self, **labels) -> None:
        if labels or self.embedded_submetrics_enabled:
            return super()._handle(**labels)

        mono_time = time.monotonic()
        self.last_clicks.append(time.monotonic())
        try:
            while self.last_clicks[0] <= mono_time - self.cutoff_period:
                self.last_clicks.popleft()
        except IndexError:
            pass

    def to_json(self) -> tp.List[int]:
        if self.embedded_submetrics_enabled:
            return super().to_json()

        count_map = [0] * len(self.time_unit_vectors)
        mono_time = time.monotonic()
        time_unit_vectors = [mono_time - v for v in self.time_unit_vectors]

        for v in self.last_clicks:
            for index, cutoff in enumerate(time_unit_vectors):
                if v >= cutoff:
                    count_map[index] += 1

        output = []
        for time_unit, count in zip(self.time_unit_vectors, count_map):
            k = super().to_json()
            k['period'] = time_unit
            k['_'] = count
            output.append(k)

        return output



"""
Routines for manipulating the JSON emitted by the metrics
"""
import copy
import logging
from satella.coding import for_argument

logger = logging.getLogger(__name__)


@for_argument(copy.copy, copy.copy)
def are_equal(tree1, tree2) -> bool:
    tree1.pop('_', None)
    tree2.pop('_', None)
    return tree1 == tree2


def is_leaf_node(tree: dict) -> bool:
    for v in tree.values():
        if isinstance(v, (dict, list, tuple)):
            return False
    return True


def annotate_every_leaf_node_with_labels(tree, labels):
    """
    Add given labels to every leaf node in the tree.

    :param tree: tree to modify in-place
    :param labels: dictionary of labels to add
    :return: tree
    """
    if isinstance(tree, list):
        return [annotate_every_leaf_node_with_labels(q, labels) for q in tree]

    if is_leaf_node(tree):
        tree.update(labels)
        return tree

    for k in tree.keys():
        tree[k] = annotate_every_leaf_node_with_labels(tree[k], labels)

    return tree


@for_argument(copy.copy, copy.copy)
def update(tree1, tree2):
    """
    Merge two dictionaries generated by Metric.to_json

    Values from tree2 take precendence

    :param tree1: a dictionary, generated by Metric.to_json
    :param tree2: a dictionary to update with, generated by Metric.to_json
    :return: a new dictionary, resulting from merging tree1 with tree2
    """

    if not isinstance(tree1, (dict, list)) and not isinstance(tree2, (dict, list)):
        return tree2

    assert type(tree1) == type(tree2), 'Cannot merge different types!'

    if isinstance(tree1, list) and isinstance(tree2, list):
        output_list = tree1
        for v_b in copy.copy(tree2):
            tree2.remove(v_b)
            for i, v_a in enumerate(tree1):
                if are_equal(v_b, v_a):
                    output_list[i] = v_b
                    break
        output_list.extend(tree2)
        return output_list

    return_tree = {}
    for k, v in tree1.items():
        v_b = tree2.pop(k, None)
        if v_b is None:
            return_tree[k] = v
        else:
            return_tree[k] = update(v, v_b)
    for k, v in tree2.items():
        return_tree[k] = v

    return return_tree

